///////////////////////////////////////////////
// 
//  Keypad Trainer Turbo Ludum Dare 48h #26 compo entry ("minimalism")
//
//  Jari Komppa aka Sol 
//  http://iki.fi/sol
// 
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
//
///////////////////////////////////////////////

#ifdef EMSCRIPTEN
#include <emscripten.h>
#endif

#include <SDL.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// Let's run physics at 100hz..
#define PHYSICS_MS 10

// The states game can be in.
enum States
{
	STATE_START, // start screen
	STATE_GAME,  // ingame
	STATE_END    // game over, man, game over..
};

// For physics loop..
int gLastTick = 0;

// Input char ('0' - '9', backspace(8) or space(32))
int gChar;

// Player's input buffer. Basically should only have one char in it, unless
// player mistypes.
char gCharBuf[22] = {0};

// Target string buffer. When this fills, it's game over.
char gTgtBuf[22] = {0};

// Milliseconds until the next character is added to buffer.
int gToNextChar = 0;

// How long we'll wait until next characters. Re-calculated constantly using gCPS.
int gNextCharPause = 1000;

// Characters per second. Multiplied by 100.
int gCPS = 100;

// Milliseconds until the next gCPS update. We increase CPS every second.
int gNextCPS = 1000;

// Offset of the target string, for smooth movement.
float gTgtX = 640;

// Game state.
int gState = 0;

// Pointer to the screen surface
SDL_Surface *gScreen;

// Pointer to our font.
SDL_Surface *gBigFont;

// Draw one character. Fixed-width. 32x32.
void blitchar32(int x, int y, char character)
{
	SDL_Rect src, dst;
	src.w = 32;
	src.h = 32;
	src.x = 0;
	src.y = 32 * character;
	dst.w = 32;
	dst.h = 32;
	dst.x = x;
	dst.y = y;
	SDL_BlitSurface(gBigFont, &src, gScreen, &dst);
}

// Draw a whole string. Whoo-ee!
void blitstring32(int x, int y, const char *text)
{
	while (text && *text)
	{
		blitchar32(x, y, *text-32);
		text++;
		x += 32;
	}
}

// Draw a filled rectangle.
void blitrect(int x, int y, int w, int h, int color)
{
/*
	SDL_Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;
	SDL_FillRect(gScreen, &r, color);
*/

	// emscripten doesn't seem to support SDL_FillRect #$@%&!!!

	SDL_LockSurface(gScreen);
	int i, j;
	for (j = 0; j < h; j++)
	{
		int * p = (int*)(gScreen->pixels) + (y+j) * gScreen->pitch / 4 + x;
		for (i = 0; i < w; i++)
		{
			*p = color;
			p++;
		}
	}
	SDL_UnlockSurface(gScreen);
}



// Resets the globals to initial values
void init(int tick)
{
	gCharBuf[0] = 0;
	gTgtBuf[0] = 0;

	gToNextChar = 0;
	gNextCharPause = 1000;
	gCPS = 100;
	gNextCPS = 1000;

	gTgtX = 640;

	gLastTick = tick;
}


// Main rendering function, with some "physics" embedded.
void reindeer(int tick)
{
	// Yeah yeah, I could keep track of the string lengths, but
	// we have all this CPU power to waste.. right?
	int tgtlen = strlen(gTgtBuf);
	int charbuflen = strlen(gCharBuf);

	// Chear screen.
	blitrect(0,0,640,480,0xff000000);

	// Draw the title.
	blitstring32(0,2,"Keypad Trainer Turbo");

	// Cycle through a few messages below the title..
	switch ((tick >> 12) & 3)
	{
	case 0:
		blitstring32(96,36,"By Jari Komppa");
		break;
	case 1:            
		blitstring32(32,36,"for Ludum Dare #26");
		break;
	case 2:
		blitstring32(48,36,"Copyright(c) 2013");
		break;
	case 3:
		blitstring32(32,36,"http://iki.fi/sol/");
		break;
	}

	// If we're in the ingame state, play the game..
	if (gState == STATE_GAME)
	{
		// "Physics" loop. Do something at 100hz..
		while (gLastTick < tick)
		{
			gToNextChar -= PHYSICS_MS;
			gNextCPS -= PHYSICS_MS;

			// ..like increase CPS..
			if (gNextCPS < 0)
			{
				gCPS += 4;
				gNextCPS = 1000;
			}

			// ..and add random numbers to the target buf..
			if (gToNextChar <= 0)
			{
				gToNextChar = gNextCharPause;
				gNextCharPause = 100000 / gCPS;
				tgtlen = strlen(gTgtBuf);
				if (tgtlen < 21)
				{
					gTgtBuf[tgtlen] = (rand() % 10) + '0';
					gTgtBuf[tgtlen+1] = 0;
				}
			}
			gLastTick += PHYSICS_MS;     

			// ..aand smoothly move the target string on screen
			gTgtX = (gTgtX * 7 + ((640 - tgtlen * 32))) / 8;
		}
    
		// if we have an input char, handle it.
		// (we probably should do this outside the physics loop, but, meh)
		if (gChar != -1)
		{
			// backspace is special..
			if (gChar == 8)
			{
				if (charbuflen > 0)
				{
					gCharBuf[charbuflen-1] = 0;
				}
			}
			else
			if (gChar == 32)
			{
				// don't do anything with spaces
			}
			else
			{
				// if there's room in the char buffer, add to it
				if (charbuflen < 21)
				{
					gCharBuf[charbuflen] = gChar;
					gCharBuf[charbuflen+1] = 0;
				}
			}
			
			// clear the input char so we don't use it more than once by accident
			gChar = -1;
		}

		// Hey look, there's something in the char buffer!
		if (gCharBuf[0] != 0)
		{
			// Does it match with the target?
			if (gCharBuf[0] == gTgtBuf[0])
			{
				// Yay! Remove first entry from both buffers (moving everything
				// in said buffers by one byte).
				// This could be handled with a ring buffer to save CPU, but
				// who cares.. it'd take more effort to code (and debug).
				int i;
				for (i = 1; i < 22; i++)
				{
					gCharBuf[i-1] = gCharBuf[i];
					gTgtBuf[i-1] = gTgtBuf[i];
				}
				// Move the target string by one character (so the string doesn't
				// move on screen when we remove one character from it)
				gTgtX+=32;
			}
		}

		// Missed too many characters, game over.
		if (tgtlen >= 21)
		{
			gState = STATE_END;
		}

		// Passed halfway point? Blink warning text.
		if (tgtlen > 10 && ((tick >> 7) & 3))
		{
			blitstring32(176, 120, "Warning!!");
		}
		
		// Helpful string.
		blitstring32(160, 240 - 40, "Type this:");
		
		// If we have a target character, display it..
		if (gTgtBuf[0])
		{
			blitchar32(320-16,240,gTgtBuf[0]-32);
		}

		// If we've missed characters, draw help and the chars.
		if (gCharBuf[0])
		{
			blitstring32(80,320, "Use backspace!");
			blitstring32((640-charbuflen*32)/2,320+34,gCharBuf);
		}
	}
	else
	if (gState == STATE_END)
	{
		// Let the player know the sad state of events.
		blitstring32(160, 160, "Game over.");

		// And blink the restart hint.
		if ((tick >> 7) & 7)
		{
			blitstring32(80, 360, "Space restarts.");
		}

		// If player has hit space, restart.
		if (gChar == 32)
		{
			gChar = -1;
			int x = gCPS;
			init(tick);
			// Don't start the speed from beginning, that'd be BORING
			gCPS = x / 1.75f;
			gState = STATE_GAME;
		}
	}
	else
	if (gState == STATE_START)
	{
		// Blink the start hint.
		if ((tick >> 7) & 7)
		{
			blitstring32(80, 240, "Space to start.");
		}

		// If player has hit space, start the game.
		if (gChar == 32)
		{
			gChar = -1;
			init(tick);
			gState = STATE_GAME;
			// The time it took for the player to hit space gives us
			// a nice random starting point.
			srand(SDL_GetTicks());
		}
	}

	// Draw the target string in all modes. In 'start' mode, it'll be empty.
	blitstring32(gTgtX, 240 + 40, gTgtBuf);

	// Finally, draw the CPS. It works as a convenient score as well.
	char temp[64];
	sprintf(temp, "cps:%3.3f", 1000.0f / gNextCharPause);
	blitstring32(160, 480 - 32, temp);
    
	// Tell SDL to update the screen.
    SDL_UpdateRect(gScreen, 0, 0, 640, 480);    
}


// The main loop!
void mainloop()
{
	// Get a tick.
	int tick = SDL_GetTicks();

	// Do some rendering..
    reindeer(tick);

	// Check for input
	gChar = -1;
    SDL_Event event;
    while ( SDL_PollEvent(&event) ) 
    {
        switch (event.type) 
        {
		case SDL_KEYDOWN:
			// Handle top-row keys, as well as keypad in both numlock modes.
			// We also want backspace and space.
			switch (event.key.keysym.sym)
			{		
				case SDLK_KP1:
				case 49:
				case 35:
					// out 1
					gChar = '1';
					break;
				case SDLK_KP2:
				case 50:
				case 1105:
					// out 2
					gChar = '2';
					break;
				case SDLK_KP3:
				case 51:
				case 1102:
					// out 3
					gChar = '3';
					break;
				case SDLK_KP4:
				case 52:
				case 1104:
					// out 4
					gChar = '4';
					break;
				case SDLK_KP5:
				case 53:
				case 12:
					// out 5
					gChar = '5';
					break;
				case SDLK_KP6:
				case 54:
				case 1103:
					// out 6
					gChar = '6';
					break;
				case SDLK_KP7:
				case 55:
				case 36:
					// out 7
					gChar = '7';
					break;
				case SDLK_KP8:
				case 56:
				case 1106:
					// out 8
					gChar = '8';
					break;
				case SDLK_KP9:
				case 57:
				case 1099:
					// out 9
					gChar = '9';
					break;
				case SDLK_KP0:
				case 48:
				case 1112:
				case 45:
					// out 0
					gChar = '0';
					break;
				case 8:
					// backspace
					gChar = 8;
					break;
				case 32:
					// Spaaaaaaaaaaaaaace!!
					gChar = 32;
					break;
				}
			break;

        case SDL_KEYUP:
			// In non-emscripten builds, it's nice to be able to quit.
            if (event.key.keysym.sym == SDLK_ESCAPE)
			{
				exit(0);
			}
            break;
        case SDL_QUIT:
            exit(0);
        }
    }
}


// Entry point
int main(int argc, char *argv[])
{      
	// We only need video, really..
    if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) 
    {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        exit(1);
    }
    
	// Cleanup for non-emscripten builds..
    atexit(SDL_Quit);
    
	// Window's good enough for us
    gScreen = SDL_SetVideoMode(640, 480, 32, SDL_SWSURFACE); 
    if (gScreen == NULL) 
    {
        printf("Unable to set 640x480 video: %s\n", SDL_GetError());
        exit(1);
    }
    

#ifdef EMSCRIPTEN
	// Browsers can load pngs. Yay!
	gBigFont = SDL_LoadBMP("bigfont.png");
#else
	// Otherwise, SDL only has built-in bmp support
	gBigFont = SDL_LoadBMP("bigfont.bmp");
#endif

	// Setup done, take the time..
    gLastTick = SDL_GetTicks();

#ifdef EMSCRIPTEN
	// Tell emscripten to call our main loop at about 60hz
	emscripten_set_main_loop(mainloop, 60, 0);
#else
	// In non-emscripten builds, just loop
	while (1)
	{
		mainloop();
	}
#endif
	// This is pretty much never reached.
    return 0;
}